; SketchyLISP Example Program
; Copyright (C) 2005,2006,2007 Nils M Holm. All rights reserved.
; See the file LICENSE of the SketchyLISP distribution
; for conditions of use.

; ---name---
; infix

; ---language---
; R5RS Scheme

; ---purpose---
; Convert arithmetic expressions from prefix to
; infix notation. Prefix expressions may contain
; variables (symbols like X), numbers (quoted
; integer numbers like -57), and these functions:
; .C + - * quotient expt.
; .B
; Functions are translated
; to these operators:
; .C + - * / ^.
; .P
; For instance,
; .B
; .C (infix '(+ -15 17))
; .B
; gives
; .B
; .C (-15 + 17).
; .P
; .C Infix
; will insert brackets ([ and ]) as required.

; ---keywords---
; INFIX function, prefix to infix, RPN, conversion

; ---args---
; X - arithmetic expression

; ---example---
; (infix '(+ 12 (* 3 (expt 4 (+ 5 6))))) => (12 + 3 * 4 ^ [ 5 + 6 ])

; (require "lib/digits.scm")
; (require "lib/list.scm")
; (require "lib/caar.scm") ; cadr
; (require "lib/caaar.scm") ; caddr
; (require "lib/append.scm")
; (require "lib/member.scm")
; (require "lib/assoc.scm")
; (require "lib/plus.scm") ; +
; (require "lib/minus.scm") ; -
; (require "lib/times.scm") ; *
; (require "lib/quotient.scm")
; (require "lib/expt.scm")

; ---code---
(define (infix x)
  (letrec
    ; Function->operator translation
    ((ops '((+ . +)
            (- . -)
            (* . *)
            (quotient . /)
            (expt . ^)))
     ; Precedence list in descending order
     (precedence
       '((high) (expt) (* quotient) (+ -) (low)))
     (function?
       (lambda (x)
         (member x '(+ - * quotient expt))))
     ; Has the operator X a higher precedence than Y?
     (higherprec?
       (lambda (x y)
         (letrec
           ((hp (lambda (x y l)
              (cond ((null? l) #f)
                ((member x (car l))
                  (not (member y (car l))))
                ((member y (car l)) #f)
                (else (hp x y (cdr l)))))))
           (hp x y precedence))))
     (parenthesize
       (lambda (x)
         (append '([) x '(]))))
     (_infix
       (lambda (x outer-op)
         (cond ((number? x) (list x))
           ((symbol? x) (list x))
           ((function? (car x))
             ; Convert prefix to infix
             (let ((sub-expr (append (_infix (cadr x) (car x))
                               (list (cdr (assoc (car x) ops)))
                               (_infix (caddr x) (car x)))))
               ; If the surrounding operation has a higher
               ; precedence, parenthesize this expression
               (cond ((higherprec? outer-op (car x))
                   (parenthesize sub-expr))
                 (else sub-expr))))
           (else (bottom (list 'syntax: x)))))))
    (_infix x 'low)))
