; SketchyLISP Example Program
; Copyright (C) 2005,2006,2007 Nils M Holm. All rights reserved.
; See the file LICENSE of the SketchyLISP distribution
; for conditions of use.

; ---name---
; prefix

; ---language---
; R5RS Scheme

; ---purpose---
; Convert arithmetic expressions in infix notation
; to S-expressions. Infix expressions are represented
; by flat lists of variables (atoms) operators (atoms)
; and numbers (sequences of atoms representing digits).
; .P
; The following operators are recognized:
; .C +
; (addition)
; .C -
; (subtraction),
; .C *
; (multiplication),
; .C /
; (division),
; .C ^
; (exponentation), and
; .C -
; (negation). Brackets are
; recoginzed as parentheses.
; .P
; For instance,
; .B
; .C (prefix '(57 * [ x + y ]))
; .B
; gives
; .B
; .C (* 57 (+ x y)).

; ---keywords---
; PREFIX function, infix to prefix, RPN, conversion

; ---args---
; X - arithmetic expression

; ---example---
; (prefix '(12 + 34 * 56 ^ [ 7 + 8 ])) => (+ 12 (* 34 (expt 56 (+ 7 8))))

; (require "lib/digits.scm")
; (require "lib/list.scm")
; (require "lib/reverse.scm")
; (require "lib/memq.scm")
; (require "lib/caar.scm") ; cadr
; (require "lib/caaar.scm") ; caadr cdadr
; (require "lib/not.scm")
; (require "lib/plus.scm") ; +
; (require "lib/minus.scm") ; -
; (require "lib/times.scm") ; *
; (require "lib/quotient.scm")
; (require "lib/expt.scm")

; ---code---
(define (prefix x)
  (letrec
    ((symbol-p
       (lambda (x)
         (memq x '(a b c d e f g h i j k l m
                   n o p q r s t u v w x y z))))
     (digit?
       (lambda (x)
         (memq x digits)))
     ; Extract a numeric value from the beginning of X.
     ; Return (N XR) where N is the value extracted and
     ; XR is the rest of X (X with N removed).
     (number
       (lambda (x r)
         (cond ((null? x)
             (list (reverse r) x))
           ((not (digit? (car x)))
             (list (reverse r) x))
           (else (number (cdr x) (cons (car x) r))))))
     ; These functions are used to extract parts
     ; of (EXPR REST) lists where EXPR is the prefix
     ; expression built so far and REST is the rest
     ; the source expression to parse
     (car-of-rest caadr)
     (cdr-of-rest cdadr)
     (expr car)
     (rest cadr)
     ; Parse factors:
     ; factor := [ sum ]
     ;   | - factor
     ;   | Number
     ;   | Symbol
     (factor
       (lambda (x)
         (cond ((null? x) (list x '()))
           ; Parse parenthesized subexpressions
           ((eq? '[ (car x))
             (let ((xsub (sum (cdr x))))
               (cond ((null? (rest xsub)) (bottom 'missing-right-paren))
                 ((eq? (car-of-rest xsub) '])
                   (list (expr xsub)
                         (cdr-of-rest xsub)))
                 (else (bottom 'missing-right-paren)))))
           ; Parse applications of unary minuses
           ((eq? '- (car x))
             (let ((fac (factor (cdr x))))
               (list (list '- (expr fac))
                     (rest fac))))
           ; Parse literal numbers
           ((digit? (car x))
             (number x '()))
           ; Parse symbols
           (else (list (car x)
                       (cdr x))))))
     ; Parse powers:
     ; power := factor
     ;   | factor ^ power
     (power
       (lambda (x)
         (let ((left (factor x)))
           (cond ((null? (rest left)) left)
             ((eq? '^ (car-of-rest left))
               (let ((right (power (cdr-of-rest left))))
                 (list (list 'expt (expr left) (expr right))
                       (rest right))))
             (else left)))))
     ; Parse terms:
     ; term := power
     ;   | power Symbol
     ;   | power * term
     ;   | power / term
     (term
       (lambda (x)
         (let ((left (power x)))
           (cond ((null? (rest left)) left)
             ((symbol-p (car-of-rest left))
               (let ((right (term (rest left))))
                 (list (list '* (expr left) (expr right))
                       (rest right))))
             ((eq? '* (car-of-rest left))
               (let ((right (term (cdr-of-rest left))))
                 (list (list '* (expr left) (expr right))
                       (rest right))))
             ((eq? '/ (car-of-rest left))
               (let ((right (term (cdr-of-rest left))))
                 (list (list 'quotient (expr left) (expr right))
                       (rest right))))
             (else left)))))
     ; Parse sums:
     ; sum := term
     ;   | term + sum
     ;   | term - sum
     (sum
       (lambda (x)
         (let ((left (term x)))
           (cond ((null? (rest left)) left)
             ((eq? '+ (car-of-rest left))
               (let ((right (sum (cdr-of-rest left))))
                 (list (list '+ (expr left) (expr right))
                       (rest right))))
             ((eq? '- (car-of-rest left))
               (let ((right (sum (cdr-of-rest left))))
                 (list (list '- (expr left) (expr right))
                       (rest right))))
             (else left))))))
    ; Pass X to the recursive descent parser consisting of
    ; SUM, TERM, POWER, FACTOR. The parsing process returns a
    ; list of the form (EXPR REST) as described above. When the
    ; REST is NIL, the entire expression could be parsed
    ; successfully.
    (let ((px (sum x)))
      (cond ((not (null? (rest px)))
          (bottom (list 'syntax 'error: (cadr px))))
        (else (expr px))))))
